#ifndef __iso_date_time_h__
#define __iso_date_time_h__

#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <time.h>
#include <string>

/*

string ms_tzname"; // "Asia/Novosibirsk"
int ms_tzoffset = 0; // from config file

tm* ms_localtime(time_t* t)
{
#	if 1
	return gmtime(*t+ms_tzoffset);
#	else
	return localtime(t);
#	endif
}

time_t ms_mktime(struct tm* m)
{
#	if 1
	return mktime(m);
#	else
	return _mkgmtime(m)-ms_tzoffset;
#	endif
}

*/

/**
	FLAGS=TZ, tzoffset=0 ==> GetLocal == GetGmt, tmx string = GMT with Z
	FLAGS=TZ, tzoffset!=0 ==> GetLocal != GetGmt, tmx string = GMT with Z
	FLAGS=TZ_OFFSET, tzoffset must be available ==> tmx string = LOCAL with TZINFO
	FLAGS=0, tzoffset must be available ==> tmx string = LOCAL w/o TZINFO

	tzoffset is always available when flags&TZ_Z == 0
	if ts is w/o tzoffset we get it from local computer time (tzoffset = _mkgmtime(&m)-mktime(&m))

	MSEC availability depends on MSEC flag
	if MSEC flag is available, msec value will be packed in tmx string
*/

class IsoDateTime
{
	time_t	ts;
	int		tzoffset;
	int		msec;
	int		flags;

public:
	enum
	{
		NTIME = -1
	};

	enum Type
	{
		GMT = 0
		, LOCAL			
	};

	enum
	{
		TZ_NONE			= 0x00
		, TZ_Z			= 0x01 //      offset
		, TZ_OFFSET		= 0x02 //  offset  
		, TZ_MASK		= 0x03

		, MSEC			= 0x04
	};

protected:
	static bool GetIsoTm(cstringref s, tm& m, int& msec, int& tzsec, int& flags)
	{
		// 0123456789012345678901234567
		// YYYY-MM-DDT01:02:03
		// YYYY-MM-DDT01:02:03.004Z		// support
		// YYYY-MM-DDT01:02:03.004		// support
		// YYYY-MM-DDT01:02:03.004+0700	// support

		// "YYYY-MM-DDT01:02:03Z" _mkgmtime->time_t
		// "YYYY-MM-DDT01:02:03" mktime->time_t
		// "YYYY-MM-DDT01:02:03+0700" _mkgmtime+0700->time_t
		// "YYYY-MM-DDT01:02:03-0700" _mkgmtime-0700->time_t

		memset(&m, 0, sizeof(m));
		msec = 0;
		tzsec = 0;
		flags = 0;

		if (s.size() < 19)
			return false;

		try
		{
			m.tm_year = boost::lexical_cast<uint32>(s.substr(0, 4))-1900;
			m.tm_mon = boost::lexical_cast<uint32>(s.substr(5, 2))-1;
			m.tm_mday = boost::lexical_cast<uint32>(s.substr(8, 2));
			m.tm_hour = boost::lexical_cast<uint32>(s.substr(11, 2));
			m.tm_min = boost::lexical_cast<uint32>(s.substr(14, 2));
			m.tm_sec = boost::lexical_cast<uint32>(s.substr(17, 2));
			m.tm_isdst = -1;

			size_t pos = 19;

			if (pos+4<=s.size() && s[pos]=='.')
			{
				msec = boost::lexical_cast<int>(s.substr(pos+1, 3));
				pos += 4;
				flags |= MSEC;
			}

			if (pos+1<=s.size() && s[pos]=='Z')
			{
				tzsec = 0;
				++pos;
				flags |= TZ_Z;
			}
			else if (pos+5<=s.size() && (s[pos]=='+' || s[pos]=='-'))
			{
				tzsec = 60*(boost::lexical_cast<int>(s.substr(pos+1, 2))*60 + boost::lexical_cast<int>(s.substr(pos+3, 2)));
				
				if (s[pos] == '-')
					tzsec = -tzsec;
				
				pos += 5;
				flags |= TZ_OFFSET;
			}

			if (pos != s.size())
				throw;
		}
		catch (...)
		{
			return false;
		}

		return true;
	}

	bool IsOk() const
	{
		return ts != NTIME;
	}

	static std::string tm2s(const tm& m)
	{
		return boost::str(boost::format("%04u-%02u-%02uT%02u:%02u:%02u") % (m.tm_year+1900) % (m.tm_mon+1) % m.tm_mday % m.tm_hour % m.tm_min % m.tm_sec);
	}

	static std::string tz2s(int tzSeconds)
	{
		return boost::str(boost::format("%s%02u%02u") % (tzSeconds>=0?"+":"-") % (tzSeconds/3600) % (tzSeconds%3600/60));
	}

public:
	static IsoDateTime time()
	{
		IsoDateTime idt;

		SYSTEMTIME st;
		GetSystemTime(&st);

		tm m;
		memset(&m, 0, sizeof(m));
		
		m.tm_year = st.wYear - 1900;
		m.tm_mon = st.wMonth - 1;
		m.tm_mday = st.wDay;
		m.tm_hour = st.wHour;
		m.tm_min = st.wMinute;
		m.tm_sec = st.wSecond;

		idt.ts = _mkgmtime(&m);
		idt.msec = st.wMilliseconds;
		return idt;
	}

public:
	IsoDateTime(time_t t = NTIME, int tzsec = 0, int _flags = 0)
		: ts(t)
		, msec(0)
		, tzoffset(tzsec)
		, flags(_flags)
	{

	}

	IsoDateTime(cstringref s, const Type type)
		: ts(NTIME)
		, msec(0)
		, tzoffset(0)
	{
		Set(s, type);
	}

	IsoDateTime(const tm& m, const Type type)
		: ts(NTIME)
		, msec(0)
		, tzoffset(0)
		, flags(0)
	{
		switch (type)
		{
		case GMT:
			SetGmt(m);
			break;
		case LOCAL:
			SetLocal(m);
			break;
		}
	}

	IsoDateTime(const time_t& t, const int _msec, const int _tzoffset, const int _flags)
		: ts(t)
		, msec(_msec)
		, tzoffset(_tzoffset)
		, flags(_flags)
	{

	}

	IsoDateTime(const tm& m, const int _msec, const int _tzoffset, const int _flags, const Type type)
		: msec(_msec)
		, tzoffset(_tzoffset)
		, flags(_flags)
	{
		Set(m, _msec, _tzoffset, _flags, type);
	}
/*
	IsoDateTime(const tmx& x)
		: ts(NTIME)
		, msec(0)
		, tzoffset(0)
		, flags(0)
	{
		Set(x);
	}
*/
	bool Set(const time_t t, const uint _msec = 0)
	{
		ts = t;
		msec = _msec;
		tzoffset = 0;
		flags = TZ_Z;
		return IsOk();
	}

	bool Set(cstringref s, const Type type)
	{
		tm m;
		int _msec, _tzoffset, _flags;
		ts = NTIME;

		if (GetIsoTm(s, m, _msec, _tzoffset, _flags))
			return Set(m, _msec, _tzoffset, _flags, type);

		return IsOk();
	}

	bool Set(const tm& m, const int _msec, const int _tzoffset, const int _flags, const Type type)
	{
		msec = _msec;
		tzoffset = _tzoffset;
		flags = _flags;

		if (flags & TZ_Z)
			ts = _mkgmtime(const_cast<tm*>(&m));
		else if (flags & TZ_OFFSET)
			ts = _mkgmtime(const_cast<tm*>(&m)) - tzoffset;
		else if (type == LOCAL)
		{
			ts = mktime(const_cast<tm*>(&m));
			tzoffset = int(_mkgmtime(const_cast<tm*>(&m)) - ts);
		}
		else
			ts = _mkgmtime(const_cast<tm*>(&m));
		
		return IsOk();
	}
/*
	bool Set(const tmx& x)
	{
		if ((x.flags & tmx::TZ_MASK) == tmx::TZ_NONE)
		{
			ts = mktime(const_cast<tm*>(&x.m));
			tzoffset = int(_mkgmtime(const_cast<tm*>(&x.m)) - ts);
			msec = x.msec;
			flags = x.flags;
		}
		else
		{
			ts = _mkgmtime(const_cast<tm*>(&x.m))-x.tzoffset;
			tzoffset = x.tzoffset;
			msec = x.msec;
			flags = x.flags; // TZ_OFFSET will be stored if available
		}

		return IsOk();
	}
*/
	bool SetGmt(const tm& m)
	{
		ts = _mkgmtime(const_cast<tm*>(&m));
		msec = 0;
		tzoffset = 0;
		flags = TZ_Z;
		return IsOk();
	}

	bool SetLocal(const tm& m)
	{
		ts = mktime(const_cast<tm*>(&m));
		msec = 0;
		tzoffset = int(_mkgmtime(const_cast<tm*>(&m)) - ts);
		flags = 0;
		return IsOk();
	}

	bool SetLocal(const tm& m, int tzSec)
	{
		ts = _mkgmtime(const_cast<tm*>(&m))-tzSec;
		tzoffset = tzSec;
		msec = 0;
		flags = 0;
		return IsOk();
	}

	tm GetLocalTm(int tzSeconds) const
	{
		if (!IsOk())
		{
			tm res;
			memset(&res, 0, sizeof(res));
			return res;
		}

		time_t t = ts + tzSeconds;
		return *gmtime(&t);
	}

	tm GetLocalTm() const
	{
		time_t t = ts + tzoffset;
		return *gmtime(&t);
	}

	tm GetGmtTm() const
	{
		return *gmtime(&ts);
	}

	tm GetTm(Type type) const
	{
		switch (type)
		{
		case GMT:
			return GetGmtTm();
		case LOCAL:
			return GetLocalTm();
		}

		tm m;
		memset(&m, 0, sizeof(m));
		return m;
	}
/*
	tmx GetLocalTmx() const
	{
		time_t t = ts + tzoffset;
		tmx x(*gmtime(&t));
		x.msec = msec;
		x.tzoffset = tzoffset;
		x.flags = flags;
		return x;
	}

	tmx GetLocalTmx(int tzSeconds, bool bPackTz = false) const
	{
		time_t t = ts + tzSeconds;
		tmx x(*gmtime(&t));
		x.tzoffset = tzSeconds;

		if (bPackTz)
			x.flags = tmx::TZ_OFFSET;

		return x;
	}

	tmx GetGmtTmx() const
	{
		tmx x(*gmtime(&ts));
		x.flags |= tmx::TZ_Z;
		return x;
	}

	tmx GetTmx(Type type) const
	{
		switch (type)
		{
		case GMT:
			return GetGmtTmx();
		case LOCAL:
			return GetLocalTmx();
		}

		return tmx();
	}

	tmx GetTmx() const
	{
		if (flags&TZ_Z)
			return rpc_time_t(*gmtime(&ts), flags, tzoffset, msec);

		time_t t = ts + tzoffset;
		return rpc_time_t(*gmtime(&t), flags, tzoffset, msec);
	}
*/
	std::string GetLocal(bool bPackTz = false) const
	{
		if (!IsOk())
			return std::string();
		
		std::string res = tm2s(GetLocalTm());

		if (bPackTz)
			res += tz2s(tzoffset);

		return res;
	}

	std::string GetLocal(int tzSeconds, bool bPackTz = false) const
	{
		if (!IsOk())
			return std::string();

		time_t t = ts + tzSeconds;
		tm m = *gmtime(&t);
		return bPackTz ? tm2s(m)+tz2s(tzSeconds) : tm2s(m);
	}

	std::string GetGmt(bool bPackTz = true) const
	{
		return !IsOk() ? std::string() : bPackTz ? tm2s(*gmtime(&ts))+"Z" : tm2s(*gmtime(&ts));
	}

	std::string GetString(Type type, bool bPackTz = true/*, bool bPackMsec = false*/) const
	{
		switch (type)
		{
		case GMT:
			return GetGmt(bPackTz);
		case LOCAL:
			return GetLocal(bPackTz);
		}

		return std::string();
	}

	string GetLocalMeta(const vector<string>& _locDows) const
	{
		static const vector<string> c_locDows = boost::assign::list_of<string>("Sun")("Mon")("Tue")("Wed")("Thu")("Fri")("Sat");
		const vector<string>& locDows = _locDows.size()==7 ? _locDows : c_locDows;

		tm m = GetLocalTm();
		return boost::str(boost::format("%s %04u-%02u-%02u %02u:%02u") % locDows[m.tm_wday%7]
			% (m.tm_year+1900) % (m.tm_mon+1) % m.tm_mday % m.tm_hour % m.tm_min);
	}

	time_t GetTs() const
	{
		return ts;
	}

	int GetTzOffset() const
	{
		return tzoffset;
	}

	operator bool () const
	{
		return IsOk();
	}

	operator time_t () const
	{
		return ts;
	}

	operator std::string() const
	{
		return GetGmt();
	}

	IsoDateTime& operator = (const time_t t)
	{
		ts = t;
		msec = 0;
		tzoffset = 0;
		flags = TZ_Z;
		return *this;
	}

	IsoDateTime& operator = (const std::string& s)
	{
		Set(s, GMT);
		return *this;
	}

	IsoDateTime& operator = (const tm& m)
	{
		SetGmt(m);
		msec = 0;
		tzoffset = 0;
		flags = TZ_Z;
		return *this;
	}

	IsoDateTime& operator = (const IsoDateTime& idt)
	{
		ts = idt.ts;
		msec = idt.msec;
		tzoffset = idt.tzoffset;
		flags = idt.flags;
		return *this;
	}

	bool operator == (const int t) const
	{
		return ts == t;
	}

	bool operator != (const int t) const
	{
		return ts != t;
	}

	bool operator < (const int t) const
	{
		return ts < t;
	}

	bool operator <= (const int t) const
	{
		return ts <= t;
	}

	bool operator > (const int t) const
	{
		return ts > t;
	}

	bool operator >= (const int t) const
	{
		return ts >= t;
	}

	bool operator == (const time_t t) const
	{
		return ts == t;
	}

	bool operator != (const time_t t) const
	{
		return ts != t;
	}

	bool operator < (const time_t t) const
	{
		return ts < t;
	}

	bool operator <= (const time_t t) const
	{
		return ts <= t;
	}

	bool operator > (const time_t t) const
	{
		return ts > t;
	}

	bool operator >= (const time_t t) const
	{
		return ts >= t;
	}

	bool operator == (const IsoDateTime& idt) const
	{
		return ts == idt.ts;
	}

	bool operator != (const IsoDateTime& idt) const
	{
		return ts != idt.ts;
	}

	bool operator < (const IsoDateTime& idt) const
	{
		return ts < idt.ts;
	}

	bool operator <= (const IsoDateTime& idt) const
	{
		return ts <= idt.ts;
	}

	bool operator > (const IsoDateTime& idt) const
	{
		return ts > idt.ts;
	}

	bool operator >= (const IsoDateTime& idt) const
	{
		return ts >= idt.ts;
	}

	IsoDateTime& operator += (const time_t t)
	{
		ts += t;
		return *this;
	}

	IsoDateTime& operator -= (const time_t t)
	{
		ts -= t;
		return *this;
	}

	IsoDateTime& operator += (const IsoDateTime& idt)
	{
		ts += idt.ts;
		return *this;
	}

	IsoDateTime& operator -= (const IsoDateTime& idt)
	{
		ts -= idt.ts;
		return *this;
	}
};

#endif // __iso_date_time_h__
