#ifndef __TValueHolder_h__
#define __TValueHolder_h__

namespace Utils
{

template<class Data> class Iterator
{
	typedef typename Data::value_type value_type;
	typedef typename Data::iterator iterator;

public:
	Iterator(Data& data):m_data(data), m_end(m_data.end()), m_it(m_data.begin())
	{
	}

	Iterator(Data& data, iterator it):m_data(data), m_end(m_data.end()), m_it(it)
	{
	}

	Iterator(Iterator& ci):m_data(ci.m_data), m_end(m_data.end()), m_it(ci.m_it)
	{
	}

	Iterator& operator ++ ()
	{
		if (*this)
			++m_it;

		return *this;
	}

	Iterator operator ++ (int)
	{
		Iterator it(*this);

		if (*this)
			++m_it;

		return it;
	}

	operator bool () const
	{
		return m_it != m_data.end();
	}

	iterator operator -> ()
	{
		return m_it;
	}

	value_type& operator * ()
	{
		return *m_it;
	}

	Iterator& operator = (const iterator& it)
	{
		m_it = it;
		return *this;
	}

	operator iterator ()
	{
		return m_it;
	}

	bool operator == (const iterator& it) const
	{
		return m_it == it;
	}

	bool operator != (const iterator& it) const
	{
		return m_it != it;
	}

private:
	Data&		m_data;
	iterator	m_end;
	iterator	m_it;
};

template<class Data> class ConstIterator
{
	typedef typename Data::value_type value_type;
	typedef typename Data::const_iterator const_iterator;

public:
	ConstIterator(const Data& data):m_data(data), m_end(m_data.end()), m_it(m_data.begin())
	{
	}

	ConstIterator(const Data& data, const_iterator it):m_data(data), m_end(m_data.end()), m_it(it)
	{
	}

	ConstIterator(const ConstIterator& ci):m_data(ci.m_data), m_end(m_data.end()), m_it(ci.m_it)
	{
	}

	ConstIterator& operator ++ ()
	{
		if (*this)
			++m_it;

		return *this;
	}

	ConstIterator operator ++ (int)
	{
		ConstIterator it(*this);

		if (*this)
			++m_it;

		return it;
	}

	operator bool () const
	{
		return m_it != m_data.end();
	}

	const_iterator operator -> ()
	{
		return m_it;
	}

	const value_type& operator * ()
	{
		return *m_it;
	}

	ConstIterator& operator = (const const_iterator& it)
	{
		m_it = it;
		return *this;
	}

	operator const_iterator ()
	{
		return m_it;
	}

	bool operator == (const const_iterator& it) const
	{
		return m_it == it;
	}

	bool operator != (const const_iterator& it) const
	{
		return m_it != it;
	}

private:
	const Data&		m_data;
	const_iterator	m_end;
	const_iterator	m_it;
};

template <class KEY, class VALUE, class KEY_IND = const KEY&> class TValueHolder
{
private:
	typedef const KEY& KEY_CREF;
	typedef KEY& KEY_REF;
	//typedef KEY_CREF KEY_IND;

	typedef VALUE* VALUE_PTR;
	typedef VALUE const * CVALUE_PTR;
	typedef VALUE& VALUE_REF; 
	typedef const VALUE& VALUE_CREF; 

private:
	typedef std::map<KEY, VALUE> Data;

public:
	typedef typename Data::key_type key_type;
	typedef typename Data::mapped_type value_type;
	typedef typename Data::iterator iterator;
	typedef typename Data::const_iterator const_iterator;
	typedef typename Data::reverse_iterator reverse_iterator;
	typedef typename Data::const_reverse_iterator const_reverse_iterator;

protected:
	iterator Find(KEY_IND key)
	{
		return m_data.find(key);
	}

	const_iterator Find(KEY_IND key) const
	{
		return m_data.find(key);
	}

	bool IsValid(iterator it) const
	{
		return it != m_data.end();
	}

	bool IsValid(const_iterator it) const
	{
		return it != m_data.end();
	}

	iterator begin()
	{
		return m_data.begin();
	}

	iterator end()
	{
		return m_data.end();
	}

	const_iterator begin() const
	{
		return m_data.begin();
	}

	const_iterator end() const
	{
		return m_data.end();
	}

	reverse_iterator rbegin()
	{
		return m_data.rbegin();
	}

	reverse_iterator rend()
	{
		return m_data.rend();
	}

	const_reverse_iterator rbegin() const
	{
		return m_data.rbegin();
	}

	const_reverse_iterator rend() const
	{
		return m_data.rend();
	}

public:
	bool IsExists(KEY_IND key) const
	{
		return IsValid(Find(key));
	}

	CVALUE_PTR GetValue(KEY_IND key) const
	{
		const_iterator it = Find(key);
		return IsValid(it) ? &it->second : 0;
	}

	VALUE_PTR GetValue(KEY_IND key)
	{
		iterator it = Find(key);
		return IsValid(it) ? &it->second : 0;
	}

	VALUE_PTR GetaValue(KEY_IND key)
	{
		VALUE_PTR pValue = GetValue(key);
		return pValue ? pValue : AddValue(key);
	}

	VALUE_PTR AddValue(KEY_IND key)
	{
		//ASSERT(!IsExists());
		pair<Data::iterator, bool> res = m_data.insert(make_pair(key, VALUE()));
		return res.second ? &res.first->second : 0;
	}

	bool SetValue(KEY_IND key, VALUE_CREF val)
	{
		VALUE_PTR pValue = GetValue(key);

		if (!pValue)
			return false;

		*pValue = val;
		return true;
	}

	bool SetaValue(KEY_IND key, VALUE_CREF val)
	{
		VALUE_PTR pValue = GetaValue(key);

		if (!pValue)
			return false;

		*pValue = val;
		return true;
	}

	void DelValue(KEY_IND key)
	{
		m_data.erase(key);
	}

	size_t size() const
	{
		return m_data.size();
	}

public:
	template<class Data, class iterator> class TIterator
	{
		typedef typename Data::mapped_type value_type;

	public:
		TIterator(Data data):m_data(data), m_it(m_data.begin())
		{
		}

		TIterator& operator ++ ()
		{
			if (this->operator bool())
				++m_it;

			return *this;
		}

		operator bool () const
		{
			return m_it != m_data.end();
		}

		value_type* operator -> ()
		{
			return &m_it->second;
		}

		const value_type* operator -> () const
		{
			return &m_it->second;
		}

		value_type& operator * ()
		{
			return m_it->second;
		}

		const value_type& operator * () const
		{
			return m_it->second;
		}

	private:
		Data		m_data;
		iterator	m_it;
	};

	typedef TIterator<TValueHolder<KEY, VALUE, KEY_IND>&, typename TValueHolder<KEY, VALUE, KEY_IND>::iterator> Iterator;
	typedef TIterator<const TValueHolder<KEY, VALUE, KEY_IND>&, typename TValueHolder<KEY, VALUE, KEY_IND>::const_iterator> ConstIterator;

private:
	Data m_data;
};

template <class T, class P> void for_each(T t, P p)
{
	for (Iterator<T> i(t); i; ++i)
		p(*i);
}

template <class C> inline string ls2s(const C& c)
{
	string r;

	for (C::const_iterator it=c.begin(); it!=c.end(); ++it)
	{
		if (r.size())
			r += ";";

		r += *it;
	}

	return r;
}

inline lstring s2ls(cstringref s)
{
	return Splitter(s, ";,");
}

inline vstring s2vs(cstringref s)
{
	return Splitter(s, ";,");
}

template <class C, class P> inline const typename C::value_type* pfind(const C& c, P p)
{
	typename C::const_iterator cit = find_if(c.begin(), c.end(), p);
	return cit!=c.end() ? &*cit : 0;
}

template <class C, class P> inline typename C::value_type* pfind(C& c, P p)
{
	typename C::iterator it = find_if(c.begin(), c.end(), p);
	return it!=c.end() ? &*it : 0;
}

} // namespace Utils

#endif // __TValueHolder_h__
