#ifndef __serializer_h__
#define __serializer_h__

#include "Types.h"
#include "StringUtils.h"
#include <string>

namespace Utils
{

class Serializer
{
public:
	typedef uint16 StrSize;
	typedef void* pvoid;
	typedef uint64 Pos;

public:
	class exception
	{
		std::string errorMessage;
	public:
		exception(const std::string& errMsg = std::string("unknown")):errorMessage(errMsg) {}
		exception(const std::string& errMsg, void* data, Pos size, Pos pos)
		{
			errorMessage = errMsg;
		}

		exception(const std::string& errMsg, void* data, Pos size,  Pos pos, Pos read_size)
		{
			errorMessage = errMsg;
		}

		const std::string& GetErrorMessage() const { return errorMessage; }
		std::string& GetErrorMessage() { return errorMessage; }
	};

protected:
	uchar* m_buf;
	Pos m_size, m_pos;

public:
	Serializer()
	{
		init(NULL, 0);
	}

	Serializer(pvoid _buf, Pos _size)
	{
		init(_buf, _size);
	}

	//

	void init(pvoid _buf, Pos _size)
	{
		m_buf = (uchar*)_buf;
		m_size = _size;
		reset();
	}

	void reset()
	{
		m_pos = 0;
	}

	//

	virtual   int8 read_int8() = 0;
	virtual  uint8 read_uint8() = 0;
	virtual  int16 read_int16() = 0;
	virtual uint16 read_uint16() = 0;
	virtual  int32 read_int32() = 0;
	virtual uint32 read_uint32() = 0;
	virtual  int64 read_int64() = 0;
	virtual uint64 read_uint64() = 0;

	virtual bool read_string(std::string& s, const size_t _sizeoflength=sizeof(StrSize)) = 0;
	virtual void read_string(char * const s, size_t s_len=-1, const size_t _sizeoflength=sizeof(StrSize)) = 0;
	virtual std::string read_string(const size_t _sizeoflength=sizeof(StrSize)) = 0;

	// write

	virtual void write(void const * const, Pos size) = 0;
	virtual void write(int8 v) = 0;
	virtual void write(uint8 v) = 0;
	virtual void write(int16 v) = 0;
	virtual void write(uint16 v) = 0;
	virtual void write(int32 v) = 0;
	virtual void write(uint32 v) = 0;
	virtual void write(int64 v) = 0;
	virtual void write(uint64 v) = 0;

	virtual void write(float v) = 0;
	virtual void write(double v) = 0;
	virtual void write(const std::string& s, const size_t _sizeoflength=sizeof(StrSize)) = 0;

	// write_*

	virtual void write_int8(int8 v) = 0;
	virtual void write_uint8(uint8 v) = 0;
	virtual void write_int16(int16 v) = 0;
	virtual void write_uint16(uint16 v) = 0;
	virtual void write_int32(int32 v) = 0;
	virtual void write_uint32(uint32 v) = 0;
	virtual void write_int64(int64 v) = 0;
	virtual void write_uint64(uint64 v) = 0;
	virtual void write_float(float v) = 0;
	virtual void write_double(double v) = 0;
	virtual void write_string(const std::string& s, const size_t _sizeoflength=sizeof(StrSize)) = 0;
	virtual void write_string(char const * const str, const size_t _sizeoflength=sizeof(StrSize)) = 0;
	virtual void write_string(char const * const str, size_t size, const size_t _sizeoflength=sizeof(StrSize)) = 0;

	//

	virtual void check(const std::string& s) const = 0;

	bool done() const
	{
		return pos() >= size();
	}

	operator bool () const
	{
		return !done();
	}

	Pos size() const
	{
		return m_size;
	}
	
	Pos pos() const
	{
		return m_pos;
	}
	void set_pos(Pos obj)
	{
		m_pos = obj;
	}

	Pos rest() const
	{
		return size()>=pos() ? size()-pos() : 0;
	}

	const pvoid data() const
	{
		return m_buf;
	}
	
	pvoid data()
	{
		return m_buf;
	}
	
	#ifdef _DEBUG
	
	pvoid get(Pos count = 0, char const * const name = "")
	{
		assert(pos() + count <= size());

		if (pos() + count > size())
			throw exception(name, m_buf + pos(), size(), pos(), count);

		return m_buf + m_pos;
	}

	const pvoid get(Pos count = 0, char const * const name = "") const
	{
		assert(pos() + count <= size());

		if (pos() + count > size())
			throw exception(name, m_buf + pos(), size(), pos(), count);

		return m_buf + m_pos;
	}

	pvoid read(Pos count = 0, char const * const name = "")
	{
		assert(pos() + count <= size());

		if (pos() + count > size())
			throw exception(name, get(), size(), pos(), count);

		Pos pos = m_pos;
		m_pos += count;
		return m_buf + pos;
	}
	
	#else

	pvoid get(Pos count = 0)
	{
		if (pos() + count > size())
			throw exception("", m_buf + pos(), size(), pos(), count);

		return m_buf + m_pos;
	}

	const pvoid get(Pos count = 0) const
	{
		if (pos() + count > size())
			throw exception("", m_buf + pos(), size(), pos(), count);

		return m_buf + m_pos;
	}

	pvoid read(Pos count = 0)
	{
		if (pos() + count > size())
			throw exception("", get(), size(), pos(), count);

		Pos pos = m_pos;
		m_pos += count;
		return m_buf + pos;
	}

	#endif // _DEBUG
};

#ifdef _DEBUG
#define _bs_read(a, b) read((a), (b))
#else
#define _bs_read(a, b) read((a))
#endif

class BufferSerializer : public Serializer
{
public:
	typedef Serializer::exception exception;

private:
	size_t read_string_size(size_t _sizeoflength)
	{
		switch (_sizeoflength)
		{
		case sizeof(uint8):
			return read_uint8();
		case sizeof(uint16):
			return read_uint16();
		case sizeof(uint32):
			return read_uint32();
		}

		return 0;
	}

public:
	BufferSerializer():Serializer()
	{
	}

	BufferSerializer(pvoid _buf, Pos _size):Serializer()
	{
		init(_buf, _size);
	}

	~BufferSerializer()
	{
	}

	// read_*

	int8 read_int8()
	{
		return *(int8*)_bs_read(sizeof(int8), "BufferSerializer::read_int8");
	}

	uint8 read_uint8()
	{
		return *(uint8*)_bs_read(sizeof(uint8), "BufferSerializer::read_uint8");
	}

	int16 read_int16()
	{
		return *(int16*)_bs_read(sizeof(int16), "BufferSerializer::read_int16");
	}

	uint16 read_uint16()
	{
		return *(uint16*)_bs_read(sizeof(uint16), "BufferSerializer::read_uint16");
	}

	int32 read_int32()
	{
		return *(int32*)_bs_read(sizeof(int32), "BufferSerializer::read_int32");
	}

	uint32 read_uint32()
	{
		return *(uint32*)_bs_read(sizeof(uint32), "BufferSerializer::read_uint32");
	}


	int64 read_int64()
	{
		return *(int64*)_bs_read(sizeof(int64), "BufferSerializer::read_int64");
	}

	uint64 read_uint64()
	{
		return *(uint64*)_bs_read(sizeof(uint64), "BufferSerializer::read_uint64");
	}

	float read_float()
	{
		return *(float*)_bs_read(sizeof(float), "BufferSerializer::read_float");
	}

	double read_double()
	{
		return *(double*)_bs_read(sizeof(double), "BufferSerializer::read_double");
	}

	bool read_string(std::string& s, const size_t _sizeoflength=sizeof(StrSize))
	{
		size_t len = read_string_size(_sizeoflength);

		if (pos() + len > size())
			return false;

		s.assign((char*)_bs_read(len, "BufferSerializer::read_string"), len);
		return true;
	}

	void read_string(char * const s, size_t s_len=-1, const size_t _sizeoflength=sizeof(StrSize))
	{
		size_t len = read_string_size(_sizeoflength);

		if (s_len!=-1 && len>s_len)
			throw exception(formatstr("BufferSerializer::read_string available buffer is small %i < %i, size()=%i, pos()=%i", s_len, len, size(), pos()));

		memcpy(s, _bs_read(len, "BufferSerializer::read_string"), len);
	}

	std::string read_string(const size_t _sizeoflength=sizeof(StrSize))
	{
		std::string s;
		read_string(s, _sizeoflength);
		return s;
	}

	std::wstring read_wstring(const size_t _sizeoflength=sizeof(StrSize))
	{
		std::wstring ws;
		read_wstring(ws, _sizeoflength);
		return ws;
	}

	void read_wstring(std::wstring& ws, const size_t _sizeoflength=sizeof(StrSize))
	{
		size_t len = read_string_size(_sizeoflength);
		ws.assign((wchar_t*)_bs_read(len, "BufferSerializer::read_wstring"), len/sizeof(std::wstring::value_type));
	}

	// write
	void write(void const * const data, Pos size)
	{
		memcpy(_bs_read(size, "BufferSerializer::write"), data, (size_t)size);
	}

	void write(int8 v)
	{
		write_int8(v);
	}

	void write(uint8 v)
	{
		write_uint8(v);
	}

	void write(int16 v)
	{
		write_int16(v);
	}

	void write(uint16 v)
	{
		write_uint16(v);
	}

	void write(int32 v)
	{
		write_int32(v);
	}

	void write(uint32 v)
	{
		write_uint32(v);
	}

	void write(int64 v)
	{
		write_int64(v);
	}

	void write(uint64 v)
	{
		write_uint64(v);
	}

	void write(float v)
	{
		write_float(v);
	}

	void write(double v)
	{
		write_double(v);
	}

	void write(const std::string& s, const size_t _sizeoflength=sizeof(StrSize))
	{
		write_string(s, _sizeoflength);
	}

	// write_*

	void write_int8(int8 v)
	{
		*(int8*)_bs_read(sizeof(int8), "BufferSerializer::write_int8") = *(int8*)&v;
	}

	void write_uint8(uint8 v)
	{
		*(uint8*)_bs_read(sizeof(uint8), "BufferSerializer::write_uint8") = *(uint8*)&v;
	}

	void write_int16(int16 v)
	{
		*(int16*)_bs_read(sizeof(int16), "BufferSerializer::write_int16") = *(int16*)&v;
	}

	void write_uint16(uint16 v)
	{
		*(uint16*)_bs_read(sizeof(uint16), "BufferSerializer::write_uint16") = *(uint16*)&v;
	}

	void write_int32(int32 v)
	{
		*(int32*)_bs_read(sizeof(int32), "BufferSerializer::write_int32") = *(int32*)&v;
	}

	void write_uint32(uint32 v)
	{
		*(uint32*)_bs_read(sizeof(uint32), "BufferSerializer::write_uint32") = *(uint32*)&v;
	}

	void write_int64(int64 v)
	{
		*(int64*)_bs_read(sizeof(int64), "BufferSerializer::write_int64") = *(int64*)&v;
	}

	void write_uint64(uint64 v)
	{
		*(uint64*)_bs_read(sizeof(uint64), "BufferSerializer::write_uint64") = *(uint64*)&v;
	}

	void write_float(float v)
	{
		*(float*)_bs_read(sizeof(float), "BufferSerializer::write_float") = *(float*)&v;
	}

	void write_double(double v)
	{
		*(double*)_bs_read(sizeof(double), "BufferSerializer::write_double") = *(double*)&v;
	}

	void write_string(const std::string& s, const size_t _sizeoflength=sizeof(StrSize))
	{
		write_string(s.data(), s.size(), _sizeoflength);
	}

	void write_string(char const * const str, const size_t _sizeoflength=sizeof(StrSize))
	{
		write_string(std::string(str), _sizeoflength);
	}

	void write_string(char const * const str, size_t _size, const size_t _sizeoflength=sizeof(StrSize))
	{
		switch (_sizeoflength)
		{
		case sizeof(uint8):
			write_uint8((uint8)_size);
			break;
		case sizeof(uint16):
			write_uint16((uint16)_size);
			break;
		case sizeof(uint32):
			write_uint32((uint32)_size);
			break;
		default:
			return;
		}

		memcpy(_bs_read((Pos)_size, "BufferSerializer::write_string"), str, _size);
	}

	//

	void check(const std::string& s) const
	{
#ifdef _DEBUG
		if (pos() != size())
			printf("BufferSerializer::size() error in parsing %s\n", s.c_str());
#endif
	}
};

#undef _bs_read

} // namespace Serializer

#endif
