#ifndef __vcalendar_h__
#define __vcalendar_h__

#include <vector>
#include <string>
#include <set>
#include <map>

#include <boost/spirit.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/assign.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>

#include "IsoDateTime.h"

bool VCal2EventInfo(string& method, cstringref vcal);

class vcalendar_record
{
public:
	struct elem
	{
		std::string id;
		std::vector<std::string> values;
		std::vector<std::pair<std::string, std::vector<std::string> > > attrs;

		void clear()
		{
			id.clear();
			values.clear();
			attrs.clear();
		}
	};

	typedef std::vector<elem> unit;

private:
	class add_id
	{
		vcalendar_record* const ptr;
	public:
		add_id(vcalendar_record* const _ptr):ptr(_ptr) {}
		void operator () (char const * const b, char const * const e) const
		{
			ptr->u.push_back(elem());
			ptr->u.back().id.assign(b, e);
		}
	};

	class add_value
	{
		vcalendar_record* const ptr;
	public:
		add_value(vcalendar_record* const _ptr):ptr(_ptr) {}
		void operator () (char const * const b, char const * const e) const
		{
			ptr->u.back().values.push_back(string(b, e));
		}
	};

	class add_param
	{
		vcalendar_record* const ptr;
	public:
		add_param(vcalendar_record* const _ptr):ptr(_ptr) {}
		void operator () (char const * const b, char const * const e) const
		{
			ptr->u.back().attrs.push_back(make_pair(std::string(b, e), std::vector<std::string>()));
		}
	};

	class add_param_value
	{
		vcalendar_record* const ptr;
	public:
		add_param_value(vcalendar_record* const _ptr):ptr(_ptr) {}
		void operator () (char const * const b, char const * const e) const
		{
			ptr->u.back().attrs.back().second.push_back(std::string(b, e));
		}
	};

	struct vcalendar_parser : public boost::spirit::grammar<vcalendar_parser>
	{
		vcalendar_parser(vcalendar_record* const _ptr):ptr(_ptr) {}

		template <typename ScannerT> struct definition
		{
			boost::spirit::rule<ScannerT>
				ALPHA
				, DIGIT
				, CR
				, LF
				, CRLF
				, SP
				, HTAB
				, WSP
				, DQUOTE
				, UTF8_2
				, UTF8_3
				, UTF8_4
				, UTF8_tail;

			boost::spirit::rule<ScannerT>
				content
				, contentline
				, name          
				, iana_token    
				, x_name        
				, vendorid      
				, param         
				, param_name    
				, param_value   
				, paramtext     
				, value         
				, quoted_string 
				, QSAFE_CHAR    
				, SAFE_CHAR     
				, VALUE_CHAR    
				, NON_US_ASCII  
				, CONTROL;
			
			boost::spirit::rule<ScannerT> const& start() const { return content; }

			definition(vcalendar_parser const& self)
			{
				using namespace boost::spirit;

				/*
				// This ABNF is just a general definition for an initial parsing
				// of the content line into its property name, parameter list,
				// and value string

				// When parsing a content line, folded lines MUST first
				// be unfolded according to the unfolding procedure
				// described above.  When generating a content line, lines
				// longer than 75 octets SHOULD be folded according to
				// the folding procedure described above.

				contentline   = name *(";" param ) ":" value CRLF
				name          = iana-token / x-name
				iana-token    = 1*(ALPHA / DIGIT / "-") // iCalendar identifier registered with IANA
				x-name        = "X-" [vendorid "-"] 1*(ALPHA / DIGIT / "-") // Reserved for experimental use.
				vendorid      = 3*(ALPHA / DIGIT) // Vendor identification
				// Each property defines the specific ABNF for the parameters
				// allowed on the property.  Refer to specific properties for
				// precise parameter ABNF.
				param         = param-name "=" param-value *("," param-value)
				param-name    = iana-token / x-name
				param-value   = paramtext / quoted-string
				paramtext     = *SAFE-CHAR
				value         = *VALUE-CHAR
				quoted-string = DQUOTE *QSAFE-CHAR DQUOTE
				QSAFE-CHAR    = WSP / %x21 / %x23-7E / NON-US-ASCII // Any character except CONTROL and DQUOTE
				SAFE-CHAR     = WSP / %x21 / %x23-2B / %x2D-39 / %x3C-7E / NON-US-ASCII // Any character except CONTROL, DQUOTE, ";", ":", ","
				VALUE-CHAR    = WSP / %x21-7E / NON-US-ASCII // Any textual character
				NON-US-ASCII  = UTF8-2 / UTF8-3 / UTF8-4 // UTF8-2, UTF8-3, and UTF8-4 are defined in [RFC3629]
				CONTROL       = %x00-08 / %x0A-1F / %x7F // All the controls except HTAB
				*/

				ALPHA = range_p('\x41', '\x5A') | range_p('\x61', '\x7A');
				DIGIT = range_p('\x30', '\x39');
				CR = ch_p('\xD');
				LF = ch_p('\xA');
				CRLF = CR >> LF;
				SP = ch_p('\x20');
				HTAB = ch_p('\x9');
				WSP = SP | HTAB;
				DQUOTE = ch_p('\x22');
				UTF8_2 = range_p('\xC2', '\xDF') >> UTF8_tail;
				UTF8_3 = (ch_p('\xE0') >> range_p('\xA0', '\xBF') >> UTF8_tail) | (range_p('\xE1', '\xEC') >> UTF8_tail >> UTF8_tail) | (ch_p('\xED') >> range_p('\x80', '\x9F') >> UTF8_tail) | (range_p('\xEE', '\xEF') >> UTF8_tail >> UTF8_tail);
				UTF8_4 = (ch_p('\xF0') >> range_p('\x90', '\xBF') >> UTF8_tail >> UTF8_tail) | (range_p('\xF1', '\xF3') >> UTF8_tail >> UTF8_tail >> UTF8_tail) | (ch_p('\xF4') >> range_p('\x80', '\x8F') >> UTF8_tail >> UTF8_tail);
				UTF8_tail = range_p('\x80', '\xBF');

				content			= *contentline;
				contentline		= name[add_id(self.ptr)] >> *(ch_p(";") >> param) >> ch_p(":") >> value[add_value(self.ptr)] >> CRLF;
				name			= iana_token | x_name;
				iana_token		= +(ALPHA | DIGIT | ch_p("-")); // iCalendar identifier registered with IANA
				x_name			= chseq_p("X-") >> !(vendorid >> ch_p("-")) >> +(ALPHA | DIGIT | ch_p("-")); // Reserved for experimental use.
				vendorid		= (ALPHA | DIGIT) >> (ALPHA | DIGIT) >> +(ALPHA | DIGIT); // Vendor identification
				param			= param_name[add_param(self.ptr)] >> ch_p("=") >> param_value[add_param_value(self.ptr)] >> *(ch_p(",") >> param_value[add_param_value(self.ptr)]);
				param_name		= iana_token | x_name;
				param_value		= paramtext | quoted_string;
				//param_value		= quoted_string | paramtext; // WORKS CORRECTLY
				paramtext		= *SAFE_CHAR;
				value			= *VALUE_CHAR;
				quoted_string	= DQUOTE >> *QSAFE_CHAR >> DQUOTE;
				QSAFE_CHAR		= WSP | ch_p('\x21') | range_p('\x23', '\x7E') | NON_US_ASCII; // Any character except CONTROL and DQUOTE
				SAFE_CHAR		= WSP | ch_p('\x21') | range_p('\x23', '\x2B') | range_p('\x2D', '\x39') | range_p('\x3C', '\x7E') | NON_US_ASCII; // Any character except CONTROL, DQUOTE, ";", ":", ","
				VALUE_CHAR		= WSP | range_p('\x21', '\x7E') | NON_US_ASCII; // Any textual character
				NON_US_ASCII	= UTF8_2 | UTF8_3 | UTF8_4; // UTF8-2, UTF8-3, and UTF8-4 are defined in [RFC3629]
				CONTROL			= range_p('\x00', '\x08') | range_p('\x0A', '\x1F') | ch_p('\x7F'); // All the controls except HTAB
			}
		};

		vcalendar_record* const ptr;
	};

	bool _parse(cstringref text)
	{
		vcalendar_parser const parser(this);
		boost::spirit::parse_info<> info = boost::spirit::parse(text.c_str(), parser, boost::spirit::nothing_p);

		if (!info.full)
			return false;

		return true;
	}

	void _clear()
	{
		u.clear();
	}

public:
	static bool check(cstringref text)
	{
		static vcalendar_record obj;
		obj._clear();
		return obj._parse(text);
	}

	static unit parse(cstringref text)
	{
		static vcalendar_record obj;
		obj._clear();
		obj._parse(text);
		return obj.u;
	}

private:
	unit	u;
};

#endif // __vcalendar_h__
